#!/usr/bin/env python3

'''
Copyright (c) 2020-2026 Children's Hospital of Philadelphia
Author: Li Fang (fangli2718@gmail.com)
              
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''


import os

import string
import sys
import shutil
import numpy as np
import pysam
from typing import List
import multiprocessing
import glob
import Levenshtein
import pyminimap2 as pymm2

from NanoRepeat import tk
from NanoRepeat.repeat_region import *
from NanoRepeat.paf import *
from NanoRepeat.split_alleles import *
from NanoRepeat.repeat_region import *

class Metainfo:
    def __init__(self, allele_id = -1, num_reads = 0, predicted_repeat_size = -1, min_repeat_size = -1, max_repeat_size = -1):
        self.allele_id = allele_id
        self.num_reads = num_reads
        self.predicted_repeat_size = predicted_repeat_size
        self.min_repeat_size = min_repeat_size
        self.max_repeat_size = max_repeat_size
    
    def init_from_repeat_size_list(self, repeat_size_list, allele_id):
        
        self.allele_id = allele_id
        self.num_reads = len(repeat_size_list)
        if len(repeat_size_list) == 0:
            self.predicted_repeat_size = -1
            self.min_repeat_size = -1 
            self.max_repeat_size = -1
            return self
        
        self.min_repeat_size = min(repeat_size_list)
        self.max_repeat_size = max(repeat_size_list)
        self.predicted_repeat_size = int(np.median(repeat_size_list) + 0.5)
        return self

    def output(self):
        return 'allele_id=%d\tnum_reads=%d\tpredicted_repeat_size=%d\tmin_repeat_size=%d\tmax_repeat_size=%d' % (self.allele_id, self.num_reads, self.predicted_repeat_size, self.min_repeat_size, self.max_repeat_size)


def extract_ref_sequence(ref_fasta_dict, repeat_region:RepeatRegion):

    MIN_ANCHOR_LEN = 10
    if repeat_region.anchor_len < MIN_ANCHOR_LEN:
        tk.eprint('WARNING: anchor_len must be >= 10!')
        repeat_region.anchor_len = 10

    chr_name = repeat_region.chrom
    if chr_name not in ref_fasta_dict:
        if 'chr' == chr_name[0:3]:
            chr_name = chr_name[3:]
        else:
            chr_name = 'chr' + chr_name
    if chr_name not in ref_fasta_dict:
        tk.eprint(f'ERROR! the following chromosome in repeat region bed file was not found in the reference fasta file: {repeat_region.chrom}')
        sys.exit()
    
    chr_seq = ref_fasta_dict[chr_name]
    chr_len = len(chr_seq)

    if repeat_region.start_pos > chr_len:
        tk.eprint(f'ERROR! the repeat start position is larger than chromosome length: {repeat_region.start_pos}. chromosome length = {chr_len}')
        sys.exit()
    
    if repeat_region.start_pos < 0:
        tk.eprint(f'ERROR! the repeat start position < 0')
        sys.exit()

    if repeat_region.end_pos > chr_len + 1:
        tk.eprint(f'ERROR! the repeat end position is larger than chromosome length: {repeat_region.end_pos}. chromosome length = {chr_len}')
        sys.exit()
    
    if repeat_region.end_pos < repeat_region.start_pos:
        tk.eprint(f'ERROR! end position is smaller than start position: {repeat_region.to_invertal()}')
        sys.exit()

    start_pos = repeat_region.start_pos - repeat_region.anchor_len
    end_pos = repeat_region.end_pos + repeat_region.anchor_len
    if start_pos < 0: 
        start_pos = 0
    
    if end_pos > chr_len:
        end_pos = chr_len
    
    repeat_region.left_anchor_seq  = chr_seq[start_pos:repeat_region.start_pos]
    repeat_region.left_anchor_len  = len(repeat_region.left_anchor_seq)

    repeat_region.right_anchor_seq = chr_seq[repeat_region.end_pos:end_pos]
    repeat_region.right_anchor_len = len(repeat_region.right_anchor_seq)

    repeat_region.mid_ref_seq      = chr_seq[repeat_region.start_pos:repeat_region.end_pos] # repeat sequence in the ref genome

    if repeat_region.left_anchor_len == 0 and repeat_region.right_anchor_len == 0:
        tk.eprint('ERROR! there is no flanking sequence around the repeat region! ')
        sys.exit(1)

    if repeat_region.left_anchor_len < MIN_ANCHOR_LEN and repeat_region.right_anchor_len < MIN_ANCHOR_LEN:
        tk.eprint(f'ERROR! both left and right flanking sequences are less than {MIN_ANCHOR_LEN} bp. The quantification might not be accurate!')
        sys.exit(1)
    
    return


def check_repeat_motif_in_ref(repeat_region:RepeatRegion, num_cpu:int):

    mid_ref_seq            = repeat_region.mid_ref_seq
    repeat_unit_seq        = repeat_region.repeat_unit_seq
    repeat_region_location = repeat_region.to_invertal()
    temp_out_dir           = repeat_region.temp_out_dir
    mid_ref_seq_file       = os.path.join(temp_out_dir, 'mid_ref_seq.fasta')
    write_seq_to_fasta(f'mid_ref_seq_{repeat_region_location}', mid_ref_seq, mid_ref_seq_file)

    pure_repeat_seq = repeat_unit_seq * int(len(mid_ref_seq) / len(repeat_unit_seq))
    edit_distance = Levenshtein.distance(pure_repeat_seq, mid_ref_seq)
    if edit_distance * 4 > min(len(pure_repeat_seq), len(mid_ref_seq)):
        repeat_region.ref_has_issue = True
        tk.eprint(f'ERROR! the repeat motif {repeat_unit_seq} in ref: {repeat_region_location} is NOT found. Seq = {mid_ref_seq}')
        
    return

def write_seq_to_fasta(seq_name, seq, out_file):

    out_f = open(out_file, 'w')
    out_f.write(f'>{seq_name}\n')
    out_f.write(f'{seq}\n')
    out_f.close()

    return

def check_anchor_mapping(one_read_anchor_paf_list: List[PAF]):

    if len(one_read_anchor_paf_list) == 0:
        return False
    
    if len(one_read_anchor_paf_list) == 1:
        return True
    
    if one_read_anchor_paf_list[0].align_len < 10: 
        return False
    
    if one_read_anchor_paf_list[0].align_score > 1.5 * one_read_anchor_paf_list[1].align_score and one_read_anchor_paf_list[0].mapq > 30:
        return True
    else:
        return False

def find_anchor_locations_for1read(read_paf_list: List[PAF], repeat_region: RepeatRegion):

    if len(read_paf_list) == 0: return

    left_anchor_paf_list  = []
    right_anchor_paf_list = []
    for paf in read_paf_list:
        if paf.tname.startswith('left_anchor'):
            left_anchor_paf_list.append(paf)
        elif paf.tname.startswith('right_anchor'):
            right_anchor_paf_list.append(paf)

    left_anchor_paf_list.sort(key = lambda paf:paf.align_score, reverse=True)
    right_anchor_paf_list.sort(key = lambda paf:paf.align_score, reverse=True)
    read = Read()
    read.read_name = read_paf_list[0].qname
    read.full_read_len = read_paf_list[0].qlen
    read.both_anchors_are_good = False
    read.left_anchor_is_good = check_anchor_mapping(left_anchor_paf_list)
    read.right_anchor_is_good = check_anchor_mapping(right_anchor_paf_list)
    
    if read.left_anchor_is_good == False: return
    if read.right_anchor_is_good == False: return
    
    repeat_region_length = 0
    read.left_anchor_paf = left_anchor_paf_list[0]
    read.right_anchor_paf = right_anchor_paf_list[0]

    # Important: if strand == '-', all positions are based on the reverse strand
    if read.left_anchor_is_good and read.right_anchor_is_good and read.left_anchor_paf.strand == read.right_anchor_paf.strand:
        repeat_region_length = read.right_anchor_paf.qstart - read.left_anchor_paf.qend

    if repeat_region_length > -10:
        read.both_anchors_are_good = True
        read.dist_between_anchors = repeat_region_length
    
    if read.both_anchors_are_good == False: return

    repeat_region.read_dict[read.read_name] = read

    repeat_region.buffer_len = 100
    if read.both_anchors_are_good:
        read.core_seq_start_pos = read.left_anchor_paf.qend - repeat_region.buffer_len
        read.core_seq_end_pos   = read.right_anchor_paf.qstart + repeat_region.buffer_len
        read.mid_seq_start_pos  = read.left_anchor_paf.qend
        read.mid_seq_end_pos    = read.right_anchor_paf.qstart
        if read.core_seq_start_pos < 0: read.core_seq_start_pos = 0
        if read.core_seq_end_pos > read.full_read_len: read.core_seq_end_pos = read.full_read_len
        read.left_buffer_len = read.left_anchor_paf.qend - read.core_seq_start_pos
        read.right_buffer_len = read.core_seq_end_pos - read.right_anchor_paf.qstart
        if read.left_anchor_paf.strand == '+':
            read.strand = '+'
        else:
            read.strand = '-'

    return

def find_anchor_locations_from_paf(repeat_region: RepeatRegion, anchor_locations_paf_text: string):

    lines = anchor_locations_paf_text.strip().split('\n')
    read_paf_list = list()
    for line in lines:
        line = line.strip()
        if not line: continue

        col_list = line.strip().split('\t')
        paf = PAF(col_list)

        if len(read_paf_list) == 0 or paf.qname == read_paf_list[0].qname:
            read_paf_list.append(paf)
        else:
            find_anchor_locations_for1read(read_paf_list, repeat_region)
            read_paf_list.clear()
            read_paf_list.append(paf)

    find_anchor_locations_for1read(read_paf_list, repeat_region)

    return

def find_anchor_locations_in_reads(data_type:string, repeat_region:RepeatRegion, num_cpu:int):

    left_template_seq    = repeat_region.left_anchor_seq
    left_template_name   = 'left_anchor'

    right_template_seq   = repeat_region.right_anchor_seq
    right_template_name  = 'right_anchor'

    template_fasta_file  = os.path.join(repeat_region.temp_out_dir, 'anchors.fasta')
    repeat_region.temp_file_list.append(template_fasta_file)

    template_fasta_fp    = open(template_fasta_file, 'w')
    template_fasta_fp.write('>%s\n' % left_template_name)
    template_fasta_fp.write('%s\n' % left_template_seq)
    template_fasta_fp.write('>%s\n' % right_template_name)
    template_fasta_fp.write('%s\n' % right_template_seq)
    template_fasta_fp.close()


    preset = tk.get_preset_for_minimap2(data_type)
    cmd = f'-c -t {num_cpu} {preset} {template_fasta_file} {repeat_region.region_fq_file}'
    mm2_out, mm2_err = pymm2.main(cmd)
    anchor_locations_paf_text = mm2_out.strip()

    find_anchor_locations_from_paf(repeat_region, anchor_locations_paf_text)
    
    return

def make_core_seq_fastq(repeat_region: RepeatRegion):
    repeat_region.core_seq_fq_file = os.path.join(repeat_region.temp_out_dir, 'core_sequences.fastq')
    repeat_region.temp_file_list.append(repeat_region.core_seq_fq_file)

    repeat_region.mid_seq_fq_file = os.path.join(repeat_region.temp_out_dir, 'middle_sequences.fastq')
    repeat_region.temp_file_list.append(repeat_region.mid_seq_fq_file)

    region_fq_f   = open(repeat_region.region_fq_file, 'r')
    core_seq_fq_f = open(repeat_region.core_seq_fq_file, 'w')
    mid_seq_fq_f  = open(repeat_region.mid_seq_fq_file, 'w')
    while 1:
        line1 = region_fq_f.readline()
        line2 = region_fq_f.readline()
        line3 = region_fq_f.readline()
        line4 = region_fq_f.readline()
        if not line1: break
        if not line2: break
        if not line3: break
        if not line4: break

        read_name = line1.strip()[1:]
        if read_name not in repeat_region.read_dict: continue
        read_sequence = line2.strip()            
        if repeat_region.read_dict[read_name].strand == '-':
            read_sequence = tk.rev_comp(read_sequence)
        core_sequence = read_sequence[repeat_region.read_dict[read_name].core_seq_start_pos:repeat_region.read_dict[read_name].core_seq_end_pos]
        mid_sequence = read_sequence[repeat_region.read_dict[read_name].mid_seq_start_pos:repeat_region.read_dict[read_name].mid_seq_end_pos]

        repeat_region.read_core_seq_dict[read_name] = core_sequence
        core_seq_fq_f.write(line1)
        core_seq_fq_f.write(core_sequence + '\n')
        core_seq_fq_f.write(line3)
        core_seq_fq_f.write('0' * len(core_sequence) + '\n')


        mid_seq_fq_f.write(line1)
        mid_seq_fq_f.write(mid_sequence + '\n')
        mid_seq_fq_f.write(line3)
        mid_seq_fq_f.write('0' * len(mid_sequence) + '\n')

    region_fq_f.close()
    core_seq_fq_f.close()
    mid_seq_fq_f.close()
    return


def round1_and_round2_estimation(data_type:string, repeat_region: RepeatRegion, num_cpu: int):

    if len(repeat_region.read_dict) == 0: return

    round1_repeat_size_list = []
    for read_name in repeat_region.read_dict:
        read = repeat_region.read_dict[read_name]
        read.round1_repeat_size = float(read.dist_between_anchors)/len(repeat_region.repeat_unit_seq)
        round1_repeat_size_list.append(read.round1_repeat_size)
    
    template_repeat_size = int(max(round1_repeat_size_list) * 1.5) + 1

    if template_repeat_size < max(round1_repeat_size_list) + 10:
        template_repeat_size = int(max(round1_repeat_size_list) + 10)
    
    round1_fasta_file  = os.path.join(repeat_region.temp_out_dir, 'round1_ref.fasta')
    repeat_region.temp_file_list.append(round1_fasta_file)
    
    round1_fasta_f = open(round1_fasta_file, 'w')
    round1_fasta_f.write(f'>{template_repeat_size}\n')
    round1_fasta_f.write(f'{repeat_region.left_anchor_seq}{repeat_region.repeat_unit_seq * template_repeat_size}\n')
    round1_fasta_f.close()

    round1_paf_file = os.path.join(repeat_region.temp_out_dir, 'round1.paf')
    repeat_region.temp_file_list.append(round1_paf_file)
    
    preset = tk.get_preset_for_minimap2(data_type)
    cmd = f'-c -t {num_cpu} {preset} -f 0.0 {round1_fasta_file} {repeat_region.core_seq_fq_file}'
    mm2_out, mm2_err = pymm2.main(cmd)
    
    lines = mm2_out.strip().split('\n')
    round2_repeat_size_dict = dict()
    
    for line in lines:
        if line.strip() == "" : continue 
        col_list = line.strip().split('\t')
        
        paf = PAF(col_list)

        if paf.tstart <= len(repeat_region.left_anchor_seq) and paf.tend >= len(repeat_region.left_anchor_seq):
            read_name = paf.qname
            round2_repeat_size = float(paf.tend - len(repeat_region.left_anchor_seq))/len(repeat_region.repeat_unit_seq)
            align_score = paf.align_score
            if read_name not in round2_repeat_size_dict:
                round2_repeat_size_dict[read_name] = []
            round2_repeat_size_dict[read_name].append((align_score, round2_repeat_size))
            
    for read_name in round2_repeat_size_dict:
        round2_repeat_size_dict[read_name].sort(key = lambda x:x[0], reverse=True)
        round2_repeat_size = round2_repeat_size_dict[read_name][0][1]
        repeat_region.read_dict[read_name].round2_repeat_size = round2_repeat_size

    round2_repeat_size_file = os.path.join(repeat_region.temp_out_dir, 'round2_repeat_size.txt')
    round2_repeat_size_f = open(round2_repeat_size_file, 'w')
    for read_name in repeat_region.read_dict:
        round2_repeat_size_f.write(f'{read_name}\t{repeat_region.read_dict[read_name].round2_repeat_size}\n')
    
    round2_repeat_size_f.close()
    
    return

def get_max_read_length_from_fastq(fq_file):

    max_read_length = 0
    fq_f = open(fq_file, 'r')
    while 1:
        line = fq_f.readline()
        if not line: break
        if len(line)-1 > max_read_length:
            max_read_length = len(line)-1
    fq_f.close()

    return max_read_length

def round3_estimation_for1read(repeat_region: RepeatRegion, read):

    lines = read.round3_paf_text.split('\n')
    read_paf_list = []
    for line in lines:
        if not line: continue
        if line.strip() == '': continue
        col_list = line.strip().split('\t')
        if len(col_list) == 0: continue
        paf = PAF(col_list)
        assert paf.qname == read.read_name
        read_paf_list.append(paf)

    if len(read_paf_list) == 0: return

    read_paf_list.sort(key = lambda paf:paf.align_score, reverse = True)
    best_paf_repeat_size_list = []
    for i in range(0, len(read_paf_list)):
        if read_paf_list[i].align_score < read_paf_list[0].align_score: break
        if read_paf_list[i].tstart < repeat_region.left_anchor_len and read_paf_list[i].tlen - read_paf_list[i].tend < repeat_region.right_anchor_len and read_paf_list[i].align_score == read_paf_list[0].align_score:
            best_paf_repeat_size_list.append(int(read_paf_list[i].tname))

    if len(best_paf_repeat_size_list) > 0:
        read.round3_repeat_size = np.mean(best_paf_repeat_size_list)
    else:
        read.round3_repeat_size = read.round2_repeat_size
    return
                
def round3_estimation_from_alignment(repeat_region: RepeatRegion):

    for read_name in repeat_region.read_dict:
        read = repeat_region.read_dict[read_name]
        if repeat_region.read_dict[read_name].round2_repeat_size == None: continue
        if read.round3_paf_text == "" : continue 
        round3_estimation_for1read(repeat_region, read)
    
    return 

def round3_estimation(data_type:string, fast_mode, repeat_region:RepeatRegion, num_cpu:int):
    
    round3_align(num_cpu, fast_mode, repeat_region, data_type)

    round3_estimation_from_alignment(repeat_region)

def round3_align (num_cpu:int, fast_mode, repeat_region:RepeatRegion, data_type:string):
    
    read_id = -1
    existing_reference_fasta_set = set()

    for read_name in repeat_region.read_dict:
        read = repeat_region.read_dict[read_name]
        round2_repeat_size = read.round2_repeat_size
        if round2_repeat_size == None: continue
        read_id += 1
        
        buffer = max(15, int(round2_repeat_size * 0.05))
        if buffer > 150: buffer = 150
        
        if fast_mode:
            buffer = 15
        
        max_template_repeat_size = int(round2_repeat_size + buffer)
        min_template_repeat_size = int(round2_repeat_size - buffer)
        
        if min_template_repeat_size < 0: min_template_repeat_size = 0
        
        template_fasta_file  = os.path.join(repeat_region.temp_out_dir, f'round3_reference.{min_template_repeat_size}-{max_template_repeat_size}.fasta')

        if template_fasta_file not in existing_reference_fasta_set:
            template_fasta_fp = open(template_fasta_file, 'w')
            for repeat_size in range(min_template_repeat_size, max_template_repeat_size+1):
                template_seq = repeat_region.left_anchor_seq + repeat_region.repeat_unit_seq * repeat_size + repeat_region.right_anchor_seq
                template_fasta_fp.write('>%s\n' % repeat_size)
                template_fasta_fp.write('%s\n' % template_seq)

            template_fasta_fp.close()

            existing_reference_fasta_set.add(template_fasta_file)

        read_seq = repeat_region.read_core_seq_dict[read_name].strip()
        read_fasta_file = os.path.join(repeat_region.temp_out_dir, f'round3_input.read{read_id}.fasta')
        
        read_fasta_f = open(read_fasta_file, 'w')
        read_fasta_f.write(f'>{read_name}\n')
        read_fasta_f.write(read_seq + '\n')
        read_fasta_f.close()

        preset = tk.get_preset_for_minimap2(data_type)
        cmd = f'{preset} -f 0.0 -N 100 -c --eqx -t {num_cpu} {template_fasta_file} {read_fasta_file}'
        mm2_out, mm2_err = pymm2.main(cmd)
        read.round3_paf_text = mm2_out

    return

def remove_noisy_reads_1d(allele_list, ploidy):

    tk.eprint('NOTICE: Trying to remove noisy reads')
    allele_list.sort(key = lambda allele:allele.num_reads)
    num_removed_reads = 0
    while len(allele_list) > ploidy and len(allele_list) >= 2:
        if allele_list[0].num_reads * 1.5 <= allele_list[-ploidy].num_reads:
            num_removed_reads += allele_list[0].num_reads 
            allele_list.pop(0)
        else:
            break

    return allele_list, num_removed_reads


def split_allele_using_gmm_1d(repeat_region:RepeatRegion, ploidy, error_rate, max_mutual_overlap, max_num_components, remove_noisy_reads):

    in_fastq_file = repeat_region.region_fq_file
    out_prefix = repeat_region.out_prefix

    if ploidy < 1:
        tk.eprint('ERROR! ploidy must be >= 1 !')
        sys.exit()


    read_repeat_count_dict = dict()
    for read_name in repeat_region.read_dict:
        if repeat_region.read_dict[read_name].round3_repeat_size != None:
            read_repeat_count_dict[read_name] = repeat_region.read_dict[read_name].round3_repeat_size

    if len(read_repeat_count_dict) == 0:
        tk.eprint(f'ERROR! No reads were found for repeat region: {repeat_region.to_outfile_prefix()}')
        return

    if len(read_repeat_count_dict) == 1:
        tk.eprint(f'ERROR! No enough reads for phasing. Repeat region is: {repeat_region.to_outfile_prefix()}')
        return

    probability_cutoff = 0.95
    cov_type = 'diag'
    dimension = 1
    
    readname_list, read_repeat_count_list = remove_outlier_reads_1d(read_repeat_count_dict)
    read_repeat_count_array = np.array(read_repeat_count_list).reshape(-1, 1)

    simulated_read_repeat_count_list = simulate_reads(read_repeat_count_list, error_rate)
    simualted_read_repeat_count_array = np.array(simulated_read_repeat_count_list).reshape(-1, dimension)

    best_n_components, final_gmm = auto_GMM_1d (simualted_read_repeat_count_array, max_num_components, cov_type, max_mutual_overlap)
    
    allele_list = create_allele_list_1d(best_n_components, final_gmm, readname_list, read_repeat_count_array, read_repeat_count_dict, probability_cutoff)
    
    num_removed_reads = 0
    if remove_noisy_reads == True and len(allele_list) > ploidy:
        allele_list, num_removed_reads = remove_noisy_reads_1d(allele_list, ploidy)
        tk.eprint(f'NOTICE: There are {len(allele_list)} alleles after removing noisy reads')
    
    best_n_components = len(allele_list)
    allele_list.sort(key = lambda allele:allele.gmm_mean1)

    readinfo_dict = create_readinfo_dict_from_allele_list(allele_list, dimension)

    output_phasing_results_1d(repeat_region, allele_list)
    
    output_phased_fastq(in_fastq_file, readinfo_dict, best_n_components, out_prefix)


    repeat_region.results.num_alleles = len(allele_list)
    output_summary_file_1d(repeat_region, allele_list, num_removed_reads, out_prefix)

    if repeat_region.no_details == False:
        plot_repeat_counts_1d(readinfo_dict, allele_list, repeat_region.to_unique_id(), out_prefix)

    return
    
def extract_fastq_from_bam(input_args, in_bam_file, repeat_region:RepeatRegion, flank_dist, out_fastq_file):
    quality_shift = 33
    assert (flank_dist >= 0)
    chrom = repeat_region.chrom
    start_pos = repeat_region.start_pos - flank_dist
    end_pos = repeat_region.end_pos + flank_dist
    if start_pos < 0: start_pos = 0

    bam = pysam.AlignmentFile(in_bam_file, "rb", reference_filename=input_args.ref_fasta)
    written_reads_set = set()
    with open(out_fastq_file, 'w') as fastq:
        for read in bam.fetch(chrom, start_pos, end_pos):
            if not read.query_sequence: continue
            if read.query_sequence == None:
                continue
            if read.query_name in written_reads_set: continue
            fastq.write(f'@{read.query_name}\n{read.query_sequence}\n+\n')
            if read.query_qualities != None:
                fastq.write(''.join([chr(q + quality_shift) for q in read.query_qualities]) + '\n')
            else:
                fastq.write(chr(quality_shift + 13) * len(read.query_sequence) + '\n')
            written_reads_set.add(read.query_name)
    bam.close()
    return

def quantify1repeat_from_bam_worker(process_id, num_para_regions, num_threads_per_region, input_args, error_rate, in_bam_file, ref_fasta_dict, repeat_region_list, result_queue):
    result_list = []
    for i in range(process_id, len(repeat_region_list), num_para_regions):
        process_name = f'Process {process_id:02}'
        repeat_region = repeat_region_list[i]
        quantified_repeat_region = quantify1repeat_from_bam(process_name, num_threads_per_region, input_args, error_rate, in_bam_file, ref_fasta_dict, repeat_region)
        result_list.append(quantified_repeat_region)
    
    result_queue.put(result_list)

    return

def quantify1repeat_from_bam(process_name, num_threads_per_region, input_args, error_rate, in_bam_file, ref_fasta_dict, repeat_region):
    
    tk.eprint(f'NOTICE: [{process_name}] Quantifying repeat: {repeat_region.to_outfile_prefix()}')
    
    def _clean_and_exit(input_args, repeat_region:RepeatRegion):

        if input_args.save_temp_files == False:
            shutil.rmtree(repeat_region.temp_out_dir)

        if repeat_region.no_details == True:
            files_to_delete = glob.glob(f'{repeat_region.out_prefix}*')
            for file_path in files_to_delete:
                try:
                    if os.path.isfile(file_path):
                        os.remove(file_path)
                except Exception as e:
                    print(f"Error in deleting file {file_path}: {e}")
        
        repeat_region.get_final_output()
        return repeat_region

    if repeat_region.chrom[0:3].lower() != 'chr':
        formatted_chr_name = 'chr' + repeat_region.chrom
    else:
        formatted_chr_name = repeat_region.chrom
     
    temp_out_dir = f'{input_args.out_prefix}.NanoRepeat_temp.{repeat_region.to_outfile_prefix()}'
    out_dir = f'{input_args.out_prefix}.details/{formatted_chr_name}'
    os.makedirs(temp_out_dir, exist_ok=True)
    os.makedirs(out_dir, exist_ok=True)

    repeat_region.temp_out_dir = temp_out_dir
    repeat_region.out_prefix = f'{out_dir}/{repeat_region.to_outfile_prefix()}'
    repeat_region.anchor_len = input_args.anchor_len

    # extract reads from bam file
    repeat_region.region_fq_file = os.path.join(temp_out_dir, f'{repeat_region.to_outfile_prefix()}.fastq')
    extract_fastq_from_bam(input_args, in_bam_file, repeat_region, repeat_region.anchor_len, repeat_region.region_fq_file)
    
    fastq_file_size = os.path.getsize(repeat_region.region_fq_file)

    if fastq_file_size == 0:
        tk.eprint(f'WARNING! No reads were found in repeat region: {repeat_region.to_outfile_prefix()}')
        return _clean_and_exit(input_args, repeat_region)

    # extract ref sequence
    extract_ref_sequence(ref_fasta_dict, repeat_region)

    if input_args.no_check_repeat_motif_in_ref == False:
        check_repeat_motif_in_ref(repeat_region, num_threads_per_region)
        if repeat_region.ref_has_issue == True and input_args.save_temp_files == False:
            return _clean_and_exit(input_args, repeat_region)
    else:
        tk.eprint(f'NOTICE: [{process_name}] Skipping checking repeat motif in reference (--no_check_repeat_motif_in_ref is set)')
        
    tk.eprint(f'NOTICE: [{process_name}] Step 1: finding anchor location in reads')
    find_anchor_locations_in_reads(input_args.data_type, repeat_region, num_threads_per_region)
    
    # make core sequence fastq
    make_core_seq_fastq(repeat_region)

    tk.eprint(f'NOTICE: [{process_name}] Step 2: round 1 and round 2 estimation')
    round1_and_round2_estimation(input_args.data_type, repeat_region, num_threads_per_region)

    tk.eprint(f'NOTICE: [{process_name}] Step 3: round 3 estimation')
    round3_estimation(input_args.data_type, input_args.fast_mode, repeat_region, num_threads_per_region)

    output_repeat_size_1d(repeat_region)

    tk.eprint(f'NOTICE: [{process_name}] Step 4: phasing reads using GMM')
    split_allele_using_gmm_1d(repeat_region, input_args.ploidy, error_rate, input_args.max_mutual_overlap, input_args.max_num_components, input_args.remove_noisy_reads)
    
    return _clean_and_exit(input_args, repeat_region)


        
def nanoRepeat_bam (input_args, in_bam_file:string):

    # ont, ont_sup, ont_q20, clr, hifi

    if input_args.data_type == 'ont' or 'clr':
        error_rate = 0.07
    elif input_args.data_type == 'ont_sup':
        error_rate = 0.04
    elif input_args.data_type == 'ont_q20':
        error_rate = 0.03
    elif input_args.data_type == 'hifi':
        error_rate = 0.02
    else:
        tk.eprint(f'ERROR! unknown data type: {input_args.data_type}')
        sys.exit(1)
    
    tk.eprint(f'NOTICE: Reading repeat region file: {input_args.repeat_region_bed}')
    repeat_region_list = read_repeat_region_file(input_args.repeat_region_bed, input_args.no_details)

    tk.eprint(f'NOTICE: Reading reference fasta file: {input_args.ref_fasta}')
    ref_fasta_dict = tk.fasta_file2dict(input_args.ref_fasta)
    
    max_num_para_regions = 16
    num_para_regions = min(max_num_para_regions, len(repeat_region_list), input_args.num_cpu)
    num_threads_per_region = int(input_args.num_cpu / num_para_regions)
    
    for i in range(0, len(repeat_region_list)):
        repeat_region_list[i].index = i

    processes = []
    result_queue = multiprocessing.Queue()
    for process_id in range(num_para_regions):
        p = multiprocessing.Process(target=quantify1repeat_from_bam_worker, args=(process_id, num_para_regions, num_threads_per_region, input_args, error_rate, in_bam_file, ref_fasta_dict, repeat_region_list, result_queue))
        processes.append(p)
        p.start()

    quantified_repeat_region = []
    for p in processes:
        quantified_repeat_region.extend(result_queue.get())

    for p in processes:
        p.join()

    quantified_repeat_region.sort(key = lambda repeat_region:repeat_region.index)

    out_tsv_file = f'{input_args.out_prefix}.NanoRepeat_output.tsv'
    out_tsv_f = open(out_tsv_file, 'w')

    for i in range(0, len(quantified_repeat_region)):
        quantified_repeat_region[i].get_final_output()
        out_tsv_f.write(quantified_repeat_region[i].final_output)
    out_tsv_f.close()

    if input_args.no_details == True:
        shutil.rmtree(f'{input_args.out_prefix}.details')
        
    tk.eprint('NOTICE: Program finished.')

    return






#!/usr/bin/env python3

'''
Copyright (c) 2020-2023 Children's Hospital of Philadelphia
Author: Li Fang (fangli2718@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''


import os
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"

import sys
import random
import numpy as np
from sklearn.mixture import GaussianMixture
from scipy.stats import norm
import math
import gzip

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
from matplotlib import cm
matplotlib.rcParams['font.family'] = "sans-serif"

from NanoRepeat import tk
from NanoRepeat.repeat_region import *

class Allele:
    def __init__(self):
        self.gmm_mean1 = None
        self.gmm_mean2 = None
        self.gmm_sd1 = None
        self.gmm_sd2 = None
        self.readname_list = []
        self.repeat1_size_list = []
        self.repeat2_size_list = []
        self.repeat1_median_size = None
        self.repeat2_median_size = None
        self.probability_list = []
        self.num_reads = None
        self.allele_frequency = None
        self.confidence_list = []

        self.gmm_min1 = None
        self.gmm_min2 = None
        self.gmm_max1 = None
        self.gmm_max2 = None

class Readinfo:
    def __init__(self, readname):

        self.readname = readname
        self.label = -1
        self.repeat_size1 = -1
        self.repeat_size2 = -1
        self.confidence = 1

def simulate_reads(read_repeat_count_list, error_rate):
    simulated_read_repeat_count_list = read_repeat_count_list * 100
    for i in range(0, len(simulated_read_repeat_count_list)):
        std = error_rate * (10 + simulated_read_repeat_count_list[i])
        random_error = random.gauss(0, std)
        simulated_read_repeat_count_list[i] += random_error
    return simulated_read_repeat_count_list

def interval_has_overlap(interval1, interval2):
    start = min(interval1[1], interval2[1])
    end = max(interval1[0], interval2[0])
    if end - start <= 0: 
        return True
    else:
        return False

def get_outlier_cutoff_from_list(repeat_count_list):

    if len(repeat_count_list) == 0:
        tk.eprint('ERROR! len(repeat_count_list) == 0')
        tk.eprint('Please report this bug!')
        sys.exit(1)

    mean = np.mean(repeat_count_list)
    std = np.std(repeat_count_list)

    min_repeat_count_cutoff = mean - 3 * std
    if min_repeat_count_cutoff < 0: min_repeat_count_cutoff = 0
    max_repeat_count_cutoff = mean + 3 * std

    return min_repeat_count_cutoff, max_repeat_count_cutoff

def analysis_outlier_1d(read_repeat_count_dict):

    repeat_count_list = list()
    for readname in read_repeat_count_dict:
        repeat_count = read_repeat_count_dict[readname]
        repeat_count_list.append(repeat_count)
    min_repeat_count_cutoff, max_repeat_count_cutoff = get_outlier_cutoff_from_list(repeat_count_list)
 
    return min_repeat_count_cutoff, max_repeat_count_cutoff

def remove_outlier_reads_2d(read_repeat_joint_count_dict):

    min_count1, max_count1, min_count2, max_count2  = analysis_outlier_2d(read_repeat_joint_count_dict)

    readname_list = list() # readnames after removing outliers
    read_repeat_count_list = list()

    for readname in read_repeat_joint_count_dict:
        repeat_count1, repeat_count2 = read_repeat_joint_count_dict[readname]
        if repeat_count1 < min_count1 or repeat_count1 > max_count1: continue 
        if repeat_count2 < min_count2 or repeat_count2 > max_count2: continue 
        readname_list.append(readname)
        read_repeat_count_list.append(repeat_count1)
        read_repeat_count_list.append(repeat_count2)

    return readname_list, read_repeat_count_list

def remove_outlier_reads_1d(read_repeat_count_dict):

    min_count, max_count  = analysis_outlier_1d(read_repeat_count_dict)

    readname_list = list() # readnames after removing outliers
    read_repeat_count_list = list()

    for readname in read_repeat_count_dict:
        repeat_count = read_repeat_count_dict[readname]
        if repeat_count < min_count or repeat_count > max_count: continue 
        readname_list.append(readname)
        read_repeat_count_list.append(repeat_count)

    return readname_list, read_repeat_count_list

def analysis_outlier_2d(read_repeat_joint_count_dict):

    repeat_count1_list = list()
    repeat_count2_list = list()

    for readname in read_repeat_joint_count_dict:
        repeat_count1, repeat_count2 = read_repeat_joint_count_dict[readname]
        repeat_count1_list.append(repeat_count1)
        repeat_count2_list.append(repeat_count2)

    min_count1, max_count1 = get_outlier_cutoff_from_list(repeat_count1_list)
    min_count2, max_count2 = get_outlier_cutoff_from_list(repeat_count2_list)
    
    return min_count1, max_count1, min_count2, max_count2

def auto_GMM_1d(X, max_num_components, cov_type, max_mutual_overlap):
    
    for n in range(2, max_num_components + 1):
        gmm = GaussianMixture(n_components=n, covariance_type=cov_type, n_init=10).fit(X)
        # gmm.means_.shape = (n, 1) (n_components, n_variable)
        # gmm.covariances_.shape = (n, 1) (n_components, n_variable)
        for i in range(0, n):
            for j in range(i+1, n):
                component_i_mean = gmm.means_[i]
                component_j_mean = gmm.means_[j]
                component_i_cov  = gmm.covariances_[i]
                component_j_cov  = gmm.covariances_[j]
                
                component_i_sd = max(1.0, math.sqrt(component_i_cov))
                component_j_sd = max(1.0, math.sqrt(component_j_cov))

                a1 = 1.0-max_mutual_overlap
                a2 = max_mutual_overlap
                interval_i = (norm.isf(a1, component_i_mean, component_i_sd), norm.isf(a2, component_i_mean, component_i_sd))
                interval_j = (norm.isf(a1, component_j_mean, component_j_sd), norm.isf(a2, component_j_mean, component_j_sd))

                if interval_has_overlap(interval_i, interval_j):
                    best_num_components = n - 1
                    gmm = GaussianMixture(n_components=best_num_components, covariance_type=cov_type, n_init=10).fit(X)
                    return best_num_components, gmm

    
    best_num_components = max_num_components
    gmm = GaussianMixture(n_components=best_num_components, covariance_type=cov_type, n_init=10).fit(X)
    return best_num_components, gmm


def auto_GMM_2d(X, max_num_components, cov_type, max_mutual_overlap):
    
    for n in range(1, max_num_components+1):
        gmm = GaussianMixture(n_components=n, covariance_type=cov_type, n_init=10).fit(X)
        # gmm.means_.shape = (n, 2) (n_components, n_variable)
        # gmm.covariances_.shape = (n, 2) (n_components, n_variable)
        for i in range(0, n):
            for j in range(i+1, n):
                component_i_mean1 = gmm.means_[i][0]
                component_i_mean2 = gmm.means_[i][1]
                component_j_mean1 = gmm.means_[j][0]
                component_j_mean2 = gmm.means_[j][1]
   
                component_i_cov1 = gmm.covariances_[i][0]
                component_i_cov2 = gmm.covariances_[i][1]
                component_j_cov1 = gmm.covariances_[j][0]
                component_j_cov2 = gmm.covariances_[j][1]
                
                component_i_sd1 = max(1.0, math.sqrt(component_i_cov1))
                component_i_sd2 = max(1.0, math.sqrt(component_i_cov2))
                component_j_sd1 = max(1.0, math.sqrt(component_j_cov1))
                component_j_sd2 = max(1.0, math.sqrt(component_j_cov2))

                a1 = 1.0-max_mutual_overlap
                a2 = max_mutual_overlap
                interval_i1 = (norm.isf(a1, component_i_mean1, component_i_sd1), norm.isf(a2, component_i_mean1, component_i_sd1))
                interval_i2 = (norm.isf(a1, component_i_mean2, component_i_sd2), norm.isf(a2, component_i_mean2, component_i_sd2))
                interval_j1 = (norm.isf(a1, component_j_mean1, component_j_sd1), norm.isf(a2, component_j_mean1, component_j_sd1))
                interval_j2 = (norm.isf(a1, component_j_mean2, component_j_sd2), norm.isf(a2, component_j_mean2, component_j_sd2))
                if interval_has_overlap(interval_i1, interval_j1) and interval_has_overlap(interval_i2, interval_j2):
                    best_num_components = n - 1
                    gmm = GaussianMixture(n_components=best_num_components, covariance_type=cov_type, n_init=10).fit(X)
                    return best_num_components, gmm

    
    best_num_components = max_num_components
    gmm = GaussianMixture(n_components=best_num_components, covariance_type=cov_type, n_init=10).fit(X)
    return best_num_components, gmm

def create_allele_list_1d(best_n_components, final_gmm, readname_list, read_repeat_count_array, read_repeat_count_dict, probability_cutoff):

    read_label_list = list(final_gmm.predict(read_repeat_count_array))
    proba2darray = final_gmm.predict_proba(read_repeat_count_array)

    allele_list = []
    for i in range(0, best_n_components):
        allele = Allele()
        allele.gmm_mean1 = final_gmm.means_[i]
        allele.gmm_sd1 = math.sqrt(final_gmm.covariances_[i])
        allele_list.append(allele)

    assert len(readname_list) == len(read_label_list)
    
    for i in range(0, len(readname_list)):
        readname = readname_list[i]
        repeat_size1 = read_repeat_count_dict[readname]
        assert repeat_size1 == read_repeat_count_array[i][0]
        read_label = read_label_list[i]
        probability = proba2darray[i][read_label]
        allele_list[read_label].readname_list.append(readname)
        allele_list[read_label].repeat1_size_list.append(repeat_size1)
        allele_list[read_label].probability_list.append(probability)

    for read_label in range(0, len(allele_list)):
        allele_list[read_label].num_reads = len(allele_list[read_label].readname_list)
        if allele_list[read_label].num_reads == 0:
            allele_list[read_label].repeat1_median_size = 0
            allele_list[read_label].gmm_min1 = 0
            allele_list[read_label].gmm_max1
            continue

        allele_list[read_label].repeat1_median_size = int(np.median(allele_list[read_label].repeat1_size_list)+0.5)
        allele_list[read_label].gmm_min1 = allele_list[read_label].gmm_mean1 - 2 * allele_list[read_label].gmm_sd1
        allele_list[read_label].gmm_max1 = allele_list[read_label].gmm_mean1 + 2 * allele_list[read_label].gmm_sd1

    for read_label in range(0, len(allele_list)):
        allele_list[read_label].confidence_list = []
        for i in range(0, len(allele_list[read_label].readname_list)):
            confidence = 'HIGH'
            if allele_list[read_label].probability_list[i] < probability_cutoff: 
                confidence = 'LOW'
            if allele_list[read_label].repeat1_size_list[i] < allele_list[read_label].gmm_min1 or allele_list[read_label].repeat1_size_list[i] > allele_list[read_label].gmm_max1:
                confidence = 'LOW'
            allele_list[read_label].confidence_list.append(confidence)

    allele_list.sort(key = lambda allele:allele.num_reads)

    while allele_list[0].num_reads == 0:
        allele_list.pop(0)
    
    return allele_list

def create_allele_list_2d(best_n_components, final_gmm, readname_list, read_repeat_count_array, read_repeat_joint_count_dict, probability_cutoff):
    
    read_label_list = list(final_gmm.predict(read_repeat_count_array))
    proba2darray = final_gmm.predict_proba(read_repeat_count_array)

    allele_list = []
    for i in range(0, best_n_components):
        allele = Allele()
        allele.gmm_mean1 = final_gmm.means_[i][0]
        allele.gmm_mean2 = final_gmm.means_[i][1]
        allele.gmm_sd1 = math.sqrt(final_gmm.covariances_[i][0])
        allele.gmm_sd2 = math.sqrt(final_gmm.covariances_[i][1])
        allele_list.append(allele)

    assert len(readname_list) == len(read_label_list)
    for i in range(0, len(readname_list)):
        readname = readname_list[i]
        repeat_size1, repeat_size2 = read_repeat_joint_count_dict[readname]
        read_label = read_label_list[i]
        probability = proba2darray[i][read_label]
        allele_list[read_label].readname_list.append(readname)
        allele_list[read_label].repeat1_size_list.append(repeat_size1)
        allele_list[read_label].repeat2_size_list.append(repeat_size2)
        allele_list[read_label].probability_list.append(probability)
    
    for read_label in range(0, len(allele_list)):
        allele_list[read_label].num_reads = len(allele_list[read_label].readname_list)
        if allele_list[read_label].num_reads == 0:
            allele_list[read_label].repeat1_median_size = 0
            allele_list[read_label].repeat2_median_size = 0
            allele_list[read_label].gmm_min1 = 0
            allele_list[read_label].gmm_min2 = 0
            allele_list[read_label].gmm_max1 = 0
            allele_list[read_label].gmm_max2 = 0
            continue
        
        allele_list[read_label].repeat1_median_size = int(np.median(allele_list[read_label].repeat1_size_list)+0.5)
        allele_list[read_label].repeat2_median_size = int(np.median(allele_list[read_label].repeat2_size_list)+0.5)
        allele_list[read_label].gmm_min1 = allele_list[read_label].gmm_mean1 - 2 * allele_list[read_label].gmm_sd1
        allele_list[read_label].gmm_min2 = allele_list[read_label].gmm_mean2 - 2 * allele_list[read_label].gmm_sd2
        allele_list[read_label].gmm_max1 = allele_list[read_label].gmm_mean1 + 2 * allele_list[read_label].gmm_sd1
        allele_list[read_label].gmm_max2 = allele_list[read_label].gmm_mean2 + 2 * allele_list[read_label].gmm_sd2

    for read_label in range(0, len(allele_list)):
        allele_list[read_label].confidence_list = []
        for i in range(0, len(allele_list[read_label].readname_list)):
            confidence = 'HIGH'
            if allele_list[read_label].probability_list[i] < probability_cutoff: 
                confidence = 'LOW'
            if allele_list[read_label].repeat1_size_list[i] < allele_list[read_label].gmm_min1 or allele_list[read_label].repeat1_size_list[i] > allele_list[read_label].gmm_max1:
                confidence = 'LOW'
            if allele_list[read_label].repeat2_size_list[i] < allele_list[read_label].gmm_min2 or allele_list[read_label].repeat2_size_list[i] > allele_list[read_label].gmm_max2:
                confidence = 'LOW'
            allele_list[read_label].confidence_list.append(confidence)

    allele_list.sort(key = lambda allele:allele.num_reads)

    while allele_list[0].num_reads == 0:
        allele_list.pop(0)

    return allele_list



def create_readinfo_dict_from_allele_list(allele_list, dimension):

    if dimension not in [1, 2]:
        tk.eprint('ERROR! dimension must be 1 or 2! Please report this bug.')
        sys.exit(1)
    
    readinfo_dict = dict()
    for read_label in range(0, len(allele_list)):
        for i in range(0, len(allele_list[read_label].readname_list)):
            readname = allele_list[read_label].readname_list[i]
            readinfo = Readinfo(readname)
            readinfo.label = read_label
            readinfo.repeat_size1 = allele_list[read_label].repeat1_size_list[i]
            if dimension == 2:
                readinfo.repeat_size2 = allele_list[read_label].repeat2_size_list[i]

            readinfo.confidence = allele_list[read_label].confidence_list[i]
            readinfo_dict[readname] = readinfo

    return readinfo_dict

def output_phasing_results_1d(repeat_region:RepeatRegion, allele_list):

    if repeat_region.no_details == False:
        out_phasing_file = repeat_region.out_prefix + '.phased_reads.txt'
        out_phasing_f = open(out_phasing_file, 'w')
        header  = f'##RepeatRegion={repeat_region.to_unique_id()}\n'
        header += f'#Read_Name\tAllele_ID\tPhasing_Confidence\tRepeat_Size\n'
        out_phasing_f.write(header)
        out = ''

    for label in range(0,len(allele_list)):
        allele_id = label + 1
        allele = allele_list[label]

        for i in range(0, len(allele.readname_list)):
            readname = allele.readname_list[i]
            repeat_size1 = allele.repeat1_size_list[i]
            confidence = allele.confidence_list[i]

            if repeat_region.no_details == False:
                out += f'{readname}\t{allele_id}\t{confidence}\t{repeat_size1:.1f}\n'

            if readname not in repeat_region.results.quantified_read_dict:
                repeat_region.results.quantified_read_dict[readname] = QuantifiedRead()
                repeat_region.results.quantified_read_dict[readname].read_name = readname

            repeat_region.results.quantified_read_dict[readname].repeat_size1 = repeat_size1
            repeat_region.results.quantified_read_dict[readname].allele_id = allele_id
            repeat_region.results.quantified_read_dict[readname].phasing_confidence = confidence

    if repeat_region.no_details == False:
        out_phasing_f.write(out)
        out_phasing_f.close()

    return

def output_phasing_results_2d(allele_list, repeat1_id, repeat2_id, in_fastq_file, out_prefix):

    out_phasing_file = out_prefix + '.phased_reads.txt'
    out_phasing_f = open(out_phasing_file, 'w')
    header  = f'##Input_FASTQ={in_fastq_file}\n'
    header += f'#Read_Name\tAllele_ID\tPhasing_Confidence\t{repeat1_id}.Repeat_Size\t{repeat2_id}.Repeat_Size\n'
    out_phasing_f.write(header)
    out = ''
    for label in range(0,len(allele_list)):
        allele_id = label + 1
        allele = allele_list[label]
        for i in range(0, len(allele.readname_list)):
            readname = allele.readname_list[i]
            repeat_size1 = allele.repeat1_size_list[i]
            repeat_size2 = allele.repeat2_size_list[i]
            confidence = allele.confidence_list[i]
            out += f'{readname}\t{allele_id}\t{confidence}\t{repeat_size1:.1f}\t{repeat_size2:.1f}\n'
    
    out_phasing_f.write(out)
    out_phasing_f.close()

    return


def output_phased_fastq(in_fastq_file, readinfo_dict, num_alleles, out_prefix):

    out_allele_fastq_file_list = list()
    for label in range(0, num_alleles):
        allele_id = label + 1
        out_allele_fastq_file = out_prefix + '.allele%d.fastq' % (allele_id)
        out_allele_fastq_file_list.append(out_allele_fastq_file)

    out_allele_fastq_f_list = list()
    for i in range(0, len(out_allele_fastq_file_list)):
        out_allele_fastq_f = open(out_allele_fastq_file_list[i], 'w')
        out_allele_fastq_f_list.append(out_allele_fastq_f)
    
    if '.gz' == in_fastq_file[-3:]:
        in_fastq_f = gzip.open(in_fastq_file, 'rt')
    else:
        in_fastq_f = open(in_fastq_file, 'rt')

    while 1:
        line1 = in_fastq_f.readline()
        line2 = in_fastq_f.readline()
        line3 = in_fastq_f.readline()
        line4 = in_fastq_f.readline()

        if not line1: break
        if not line2: break
        if not line3: break
        if not line4: break

        readname = line1.strip().split()[0][1:]
        if readname not in readinfo_dict: continue
        if readinfo_dict[readname].confidence != 'HIGH': continue
    
        label = readinfo_dict[readname].label
        out_allele_fastq_f_list[label].write(line1 + line2 + line3 + line4)

    in_fastq_f.close()

    for i in range(0, len(out_allele_fastq_f_list)):
        out_allele_fastq_f_list[i].close()

    return

def output_summary_file_1d(repeat_region:RepeatRegion, allele_list, num_removed_reads, out_prefix):

    num_alleles = len(allele_list)
    if repeat_region.no_details == False:
        
        out_summray_file = out_prefix + '.summary.txt'
        filebasename = os.path.split(out_summray_file)[1]
        out_summray_f = open(out_summray_file, 'w')
        summary_info  = f'Summary_file={filebasename}\tRepeat_Region={repeat_region.to_unique_id()}'
        summary_info += f'\tMethod=GMM'
        summary_info += f'\tNum_Alleles={num_alleles}'
        summary_info += f'\tNum_Removed_Reads={num_removed_reads}'

    for label in range(0, num_alleles):
        allele_id = label + 1
        if repeat_region.no_details == False:
            summary_info += f'\tAllele{allele_id}_Num_Reads={allele_list[label].num_reads}'
            summary_info += f'\tAllele{allele_id}_Repeat_Size={allele_list[label].repeat1_median_size}'

        quantified_allele = QuantifiedAllele()
        quantified_allele.repeat_size1 = allele_list[label].repeat1_median_size
        quantified_allele.num_supp_reads = allele_list[label].num_reads
        repeat_region.results.quantified_allele_list.append(quantified_allele)

        
    if repeat_region.no_details == False:
        summary_info += '\n'
        out_summray_f.write(summary_info)
        out_summray_f.close()

    return

def output_summary_file_2d(in_fastq_file, allele_list, repeat1_id, repeat2_id, num_removed_reads, out_prefix):

    num_alleles = len(allele_list)
    out_summray_file = out_prefix + '.summary.txt'
    out_summray_f = open(out_summray_file, 'w')
    summary_info  = f'Input_FASTQ\t{in_fastq_file}\n'
    summary_info += f'Method\t2D-GMM\n'
    summary_info += f'Num_Alleles\t{num_alleles}\n'
    summary_info += f'Num_Removed_Reads\t{num_removed_reads}\n'

    for label in range(0, num_alleles):
        allele_id = label + 1
        summary_info += f'Allele{allele_id}_Num_Reads\t{allele_list[label].num_reads}\n'
        summary_info += f'Allele{allele_id}_{repeat1_id}.Repeat_Size\t{allele_list[label].repeat1_median_size}\n'
        summary_info += f'Allele{allele_id}_{repeat2_id}.Repeat_Size\t{allele_list[label].repeat2_median_size}\n'

    out_summray_f.write(summary_info)
    out_summray_f.close()

    return

def output_repeat_size_1d(repeat_region:RepeatRegion):

    if repeat_region.no_details == False:
        repeat_size_file =  f'{repeat_region.out_prefix}.repeat_size.txt'
        repeat_size_f = open(repeat_size_file, 'w')
        header  = f'##Repeat_Region={repeat_region.to_unique_id()}\n'
        header += f'#Read_Name\tRepeat_Size\n'
        repeat_size_f.write(header)
    
    for read_name in repeat_region.read_dict:
        repeat_size = repeat_region.read_dict[read_name].round3_repeat_size
        if repeat_size != None:
            if repeat_region.no_details == False:
                repeat_size_f.write(f'{read_name}\t{repeat_size:.1f}\n')
            if read_name not in repeat_region.results.quantified_read_dict:
                repeat_region.results.quantified_read_dict[read_name] = QuantifiedRead()
                repeat_region.results.quantified_read_dict[read_name].read_name = read_name
                repeat_region.results.quantified_read_dict[read_name].repeat_size1 = repeat_size

    if repeat_region.no_details == False:
        repeat_size_f.close()

    return

def output_repeat_size_2d (in_fastq_file, repeat1_id, repeat2_id, out_prefix, repeat1_count_dict, repeat2_count_dict):

    repeat_size_file = f'{out_prefix}.repeat_size.txt'
    repeat_size_f = open(repeat_size_file, 'w')
    header = f'##Input_FASTQ={in_fastq_file}\n'
    header += f'#Read_Name\t{repeat1_id}.Repeat_Size\t{repeat2_id}.Repeat_Size\n'
    repeat_size_f.write(header)

    all_readname_list = []
    for readname in repeat1_count_dict:
        all_readname_list.append(readname)
    for readname in repeat2_count_dict:
        all_readname_list.append(readname)

    all_readname_set = set(all_readname_list)
    read_repeat_joint_count_dict = dict()
    read_repeat_joint_count_list = []
    for readname in all_readname_set:
        if readname in repeat1_count_dict: 
            size1 = repeat1_count_dict[readname]
        else:
            size1 = 'N.A.'

        if readname in repeat2_count_dict:
            size2 = repeat2_count_dict[readname]
        else:
            size2 = 'N.A.'
        
        read_repeat_joint_count_dict[readname] = (size1, size2)
        tup = (readname, size1, size2)
        read_repeat_joint_count_list.append(tup)
    
    read_repeat_joint_count_list.sort(key = lambda x:x[1])
    for tup in read_repeat_joint_count_list:
        readname, size1, size2 = tup
        repeat_size_f.write(f'{readname}\t{size1:.1f}\t{size2:.1f}\n')
    repeat_size_f.close()

    tk.eprint(f'NOTICE: Repeat size file is here: {repeat_size_file}')
    return read_repeat_joint_count_dict


def plot_repeat_counts_1d(readinfo_dict, allele_list, repeat_id, out_prefix):

    hist_figure_file = out_prefix + '.hist.png'
    num_alleles = len(allele_list)
    x_list = list()

    x_2d_list = [0] * num_alleles
    for i in range(0, num_alleles):
        x_2d_list[i] = list()

    for readname in readinfo_dict:
        readinfo = readinfo_dict[readname]

        x = readinfo.repeat_size1

        x_list.append(x)

        x_2d_list[readinfo.label].append(x)

    xmin = int(min(x_list))
    xmax = int(max(x_list))+1

    if xmax - xmin <= 200:
        b1 = range(xmin - 1, xmax + 2)
    else:
        b1 = range(xmin - 1, xmax + 2, int(float(xmax - xmin)/200.0 + 0.5))

    repeat1_predicted_size_list = []
    for label in range(0, len(allele_list)):
        repeat1_predicted_size_list.append(allele_list[label].repeat1_median_size)

    plot_hist1d(x_2d_list, b1, repeat_id, repeat1_predicted_size_list, allele_list, hist_figure_file)

    return

def plot_repeat_counts_2d(readinfo_dict, allele_list, repeat1_id, repeat2_id, out_prefix):

    hist2d_figure_file = out_prefix + '.hist2d.png'
    repeat1_hist_figure_file  = out_prefix + '.%s.hist.png' % (repeat1_id)
    repeat2_hist_figure_file  = out_prefix + '.%s.hist.png' % (repeat2_id)

    num_alleles = len(allele_list)
    x_list = list()
    y_list = list()

    x_2d_list = [0] * num_alleles
    y_2d_list = [0] * num_alleles
    for i in range(0, num_alleles):
        x_2d_list[i] = list()
        y_2d_list[i] = list()

    for readname in readinfo_dict:
        readinfo = readinfo_dict[readname]

        x = readinfo.repeat_size1
        y = readinfo.repeat_size2

        x_list.append(x)
        y_list.append(y)

        x_2d_list[readinfo.label].append(x)
        y_2d_list[readinfo.label].append(y)

    xmin = int(min(x_list))
    xmax = int(max(x_list))+1
    ymin = int(min(y_list))
    ymax = int(max(y_list))+1

    if xmax - xmin <= 200:
        b1 = range(xmin - 1, xmax + 2)
    else:
        b1 = range(xmin - 1, xmax + 2, int(float(xmax - xmin)/200.0 + 0.5))

    if ymax - ymin <= 200:
        b2 = range(ymin - 1, ymax + 2)
    else:
        b2 = range(ymin - 1, ymax + 2, int(float(ymax - ymin)/200.0 + 0.5))
    
    repeat1_predicted_size_list = []
    repeat2_predicted_size_list = []
    for label in range(0, len(allele_list)):
        repeat1_predicted_size_list.append(allele_list[label].repeat1_median_size)
        repeat2_predicted_size_list.append(allele_list[label].repeat2_median_size)

    plot_hist2d(x_list, y_list, b1, b2, repeat1_id, repeat2_id, allele_list, hist2d_figure_file)
    plot_hist1d(x_2d_list, b1, repeat1_id, repeat1_predicted_size_list, allele_list, repeat1_hist_figure_file)
    plot_hist1d(y_2d_list, b2, repeat2_id, repeat2_predicted_size_list, allele_list, repeat2_hist_figure_file)

    return 

def get_1d_max_min_from_allele_list(allele_list):

    gmm_min_list = []
    gmm_max_list = []

    for allele in allele_list:
        gmm_min_list.append(allele.gmm_min1)
        gmm_max_list.append(allele.gmm_max1)

    xmin = min(gmm_min_list)
    xmax = max(gmm_max_list)

    xmax = int(((xmax * 1)/10.0 +1 )* 10)
    xmin = int(((xmin)/10.0 -1 )* 10)

    if xmin < 10: xmin = 0

    return xmin, xmax

def get_2d_max_min_from_allele_list(allele_list):

    assert(len(allele_list) > 0)

    gmm_min1_list = []
    gmm_max1_list = []
    gmm_min2_list = []
    gmm_max2_list = []

    for allele in allele_list:
        gmm_min1_list.append(allele.gmm_min1)
        gmm_min2_list.append(allele.gmm_min2)
        gmm_max1_list.append(allele.gmm_max1)
        gmm_max2_list.append(allele.gmm_max2)

    xmin = min(gmm_min1_list)
    xmax = max(gmm_max1_list)
    ymin = min(gmm_min2_list)
    ymax = max(gmm_max2_list)

    xmax = int(((xmax * 1)/10.0 +1 )* 10)
    ymax = int(((ymax * 1)/10.0 +1 )* 10)

    xmin = int(((xmin)/10.0 -1 )* 10)
    ymin = int(((ymin)/10.0 -1 )* 10)

    if xmin < 10: xmin = 0
    if ymin < 10: ymin = 0
    
    return xmin, xmax, ymin, ymax


def plot_hist1d(x_2d_list, b, repeat_id, predicted_size_list, allele_list, out_file):

    plt.figure (figsize=(6, 4))
    
    max_x = 0
    for x in x_2d_list:
        if len(x) == 0: continue
        plt.hist(x, bins = b)
        if max_x < max(x):
            max_x = max(x)

    for repeat_size in predicted_size_list:
        plt.axvline(x=repeat_size+0.5, color = 'grey', linestyle = ':')

    if len(repeat_id) > 30:
        repeat_id = repeat_id[0:30] + '...'
    
    plt.title('Repeat size distribution (%s)' % repeat_id)
    plt.xlabel('repeat size')
    plt.ylabel('number of reads')
    
    xmin, xmax = get_1d_max_min_from_allele_list(allele_list)
    
    plt.xlim(xmin, xmax)

    plt.savefig(out_file, dpi=300)
    plt.close('all')

    return

def plot_hist2d(x_list, y_list, b1, b2, repeat_id1, repeat_id2, allele_list, out_file):

    plt.figure (figsize=(8, 4))
    plt.hist2d (x_list, y_list, [b1, b2], cmap = 'binary')
    plt.colorbar()
    plt.title('2D histogram of repeat size')
    plt.xlabel('repeat size (%s)' % repeat_id1)
    plt.ylabel('repeat size (%s)' % repeat_id2)

    xmin, xmax, ymin, ymax = get_2d_max_min_from_allele_list(allele_list)
   
    '''
    xmin = min(x_list)
    xmax = max(x_list)
    ymin = min(y_list)
    ymax = max(y_list)
    '''
    
    if xmax - xmin < 20: 
        delta = (20 - (xmax - xmin))/2
        xmin -= delta
        if xmin < 0: xmin = 0
        xmax = xmin + 20
    if ymax - ymin < 20: 
        delta = (20-(ymax-ymin))/2
        ymin -= delta
        if ymin < 0: ymin = 0
        ymax = ymin + 20
    
    plt.xlim(xmin, xmax)
    plt.ylim(ymin, ymax)

    plt.savefig(out_file, dpi=300)
    plt.close('all')

    return

def scatter_plot_with_contour_2d (read_repeat_joint_count_dict, final_gmm, score_cut_off, repeat1_id, repeat2_id, allele_list, out_prefix):

    scatter_plot_file = out_prefix + '.scatter.png'

    xlabel = '%s repeat size' % (repeat1_id)
    ylabel = '%s repeat size' % (repeat2_id)

    X = list()
    Y = list()
    
    for readname in read_repeat_joint_count_dict:
        x, y = read_repeat_joint_count_dict[readname]
        X.append(x)
        Y.append(y)

    X = np.array(X)
    Y = np.array(Y)

    X, Y, Z= countxy(X, Y)
    fig, ax = plt.subplots()
    fig.set_size_inches(6, 4)

    try:
        ax.scatter(X, Y, c=Z, s=15, edgecolor='')
    except:
        ax.scatter(X, Y, c=Z, s=15, edgecolor=['none'])

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    norm = Normalize(vmin = np.min(Z), vmax = np.max(Z))
    cbar = fig.colorbar(cm.ScalarMappable(norm = norm), ax=ax)
    cbar.ax.set_ylabel('Count')

    figure_min_x, figure_max_x, figure_min_y, figure_max_y = get_2d_max_min_from_allele_list(allele_list)

    a = np.linspace(figure_min_x, figure_max_x, 400)
    b = np.linspace(figure_min_y, figure_max_y, 400)

    A, B = np.meshgrid(a, b)
    AA = np.array([A.ravel(), B.ravel()]).T
    C = final_gmm.score_samples(AA)
    C = C.reshape(A.shape)

    CS = plt.contour(A, B, C, levels=[score_cut_off], linestyles = 'dashed', colors = 'grey')

    plt.savefig(scatter_plot_file, dpi = 300)
    plt.close('all')
    return

def countxy(x, y):
    count_dict = dict()
    for i in range(0, len(x)):
        key = '%d\t%d' % (x[i], y[i])
        if key not in count_dict:
            count_dict[key] = 1
        else:
            count_dict[key] += 1
    x_list = list()
    y_list = list()
    z_list = list()
    for key in count_dict:
        x, y = key.split('\t')
        x = int(x)
        y = int(y)
        z = count_dict[key]
        x_list.append(x)
        y_list.append(y)
        z_list.append(z)
    
    return x_list, y_list, z_list


#!/usr/bin/env python3

'''
Copyright (c) 2020-2023 Children's Hospital of Philadelphia
Author: Li Fang (fangli2718@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import os
import sys
from NanoRepeat import tk

class PAF:
    def __init__(self, col_list):
        if len(col_list) < 12:
            tk.eprint('ERROR! number of columns should be >= 12 in a PAF file!')
            tk.eprint('The corrupted line is: %s' % '\t'.join(col_list))
            sys.exit()
        
        self.qname, self.qlen, self.qstart, self.qend = col_list[0:4]
        self.strand = col_list[4]
        self.tname, self.tlen, self.tstart, self.tend = col_list[5:9]
        self.n_match, self.align_len, self.mapq = col_list[9:12]

        self.qlen      = int(self.qlen)
        self.qstart    = int(self.qstart)
        self.qend      = int(self.qend)
        self.tlen      = int(self.tlen)
        self.tstart    = int(self.tstart)
        self.tend      = int(self.tend)
        self.n_match   = int(self.n_match)
        self.align_len = int(self.align_len)
        self.mapq      = int(self.mapq)
        self.is_primary= False
        self.align_score = -1
        self.cigar = ''
        for col in col_list[12:]:
            if col[0:5] == 'AS:i:': 
                self.align_score = int(col[5:])
            elif col[0:5] == 'cg:Z:':
                self.cigar = col[5:]
            elif col == 'tp:A:P':
                self.is_primary = True
            elif col == 'tp:A:S':
                self.is_primary = False

        if self.strand != '+' and self.strand != '-':
            tk.eprint('ERROR! unknown strand: %s' % self.strand)
            sys.exit()
        
        if self.strand == '-':
            new_qstart  = self.qlen - self.qend
            new_qend    = self.qlen - self.qstart
            self.qstart = new_qstart
            self.qend   = new_qend
        
        self.tscore = 0

    def output_core(self):
        return f'{self.qname}\t{self.strand}\t{self.qlen}\t{self.qstart}\t{self.qend}\t{self.tname}\t{self.tlen}\t{self.tstart}\t{self.tend}'
        

        


"""
NanoRepeat: quantification of Short Tandem Repeats (STRs) from long-read sequencing data
"""

__author__ = 'Li Fang and Kai Wang'
__credits__ = "Children's Hospital of Philadelphia"
__version__ = '1.8.3'
